Package: BiocParallel
Type: Package
Title: Bioconductor facilities for parallel evaluation
Version: 1.47.0
Authors@R: c(
    person("Jiefei", "Wang", 
        email = "jiefei0804@gmail.com", 
        role = c("aut", "cre")),
    person("Martin", "Morgan", email = "mtmorgan.bioc@gmail.com", role="aut"),
    person("Valerie", "Obenchain", role="aut"),
    person("Michel", "Lang", email="michellang@gmail.com", role="aut"),
    person("Ryan", "Thompson", email="rct@thompsonclan.org", role="aut"),
    person("Nitesh", "Turaga", role="aut"),
    person("Aaron", "Lun", role = "ctb"),
    person("Henrik", "Bengtsson", role = "ctb"),
    person("Madelyn", "Carlson", role = "ctb",
        comment = "Translated 'Random Numbers' vignette from Sweave to RMarkdown / HTML."
    ),
    person("Phylis", "Atieno", role = "ctb",
        comment = "Translated 'Introduction to BiocParallel' vignette from Sweave to Rmarkdown / HTML."
    ),
    person(
        "Sergio", "Oller", role = "ctb",
        comment = c(
            "Improved bpmapply() efficiency.",
            "ORCID" = "0000-0002-8994-1549"
        )
    ))
Description: This package provides modified versions and novel
    implementation of functions for parallel evaluation, tailored to
    use with Bioconductor objects.
URL: https://github.com/Bioconductor/BiocParallel
BugReports: https://github.com/Bioconductor/BiocParallel/issues
biocViews: Infrastructure
License: GPL-2 | GPL-3 | BSL-1.0
SystemRequirements: C++11
Depends: methods, R (>= 4.1.0)
Imports: stats, utils, futile.logger, parallel, snow, codetools
Suggests: BiocGenerics, tools, foreach, BBmisc, doParallel,
        GenomicRanges, RNAseqData.HNRNPC.bam.chr14,
        TxDb.Hsapiens.UCSC.hg19.knownGene, VariantAnnotation,
        Rsamtools, GenomicAlignments, ShortRead, RUnit, BiocStyle,
        knitr, batchtools, data.table
Enhances: Rmpi
Collate: AllGenerics.R DeveloperInterface.R prototype.R bploop.R
        ErrorHandling.R log.R bpbackend-methods.R bpisup-methods.R
        bplapply-methods.R bpiterate-methods.R bpstart-methods.R
        bpstop-methods.R BiocParallelParam-class.R bpmapply-methods.R
        bpschedule-methods.R bpvec-methods.R bpvectorize-methods.R
        bpworkers-methods.R bpaggregate-methods.R bpvalidate.R
        SnowParam-class.R MulticoreParam-class.R
        TransientMulticoreParam-class.R register.R SerialParam-class.R
        DoparParam-class.R SnowParam-utils.R BatchtoolsParam-class.R
        progress.R ipcmutex.R worker-number.R utilities.R rng.R
        bpinit.R reducer.R worker.R bpoptions.R cpp11.R
        BiocParallel-defunct.R
LinkingTo: BH (>= 1.87.0), cpp11
VignetteBuilder: knitr
RoxygenNote: 7.1.2
git_url: https://git.bioconductor.org/packages/BiocParallel
git_branch: devel
git_last_commit: ec5322d
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-05-18
NeedsCompilation: yes
Packaged: 2026-05-18 21:23:48 UTC; biocbuild
Author: Jiefei Wang [aut, cre],
  Martin Morgan [aut],
  Valerie Obenchain [aut],
  Michel Lang [aut],
  Ryan Thompson [aut],
  Nitesh Turaga [aut],
  Aaron Lun [ctb],
  Henrik Bengtsson [ctb],
  Madelyn Carlson [ctb] (Translated 'Random Numbers' vignette from Sweave
    to RMarkdown / HTML.),
  Phylis Atieno [ctb] (Translated 'Introduction to BiocParallel' vignette
    from Sweave to Rmarkdown / HTML.),
  Sergio Oller [ctb] (Improved bpmapply() efficiency., ORCID:
    <https://orcid.org/0000-0002-8994-1549>)
Maintainer: Jiefei Wang <jiefei0804@gmail.com>
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-19 04:43:42 UTC; unix
