# Generated by cpp11: do not edit by hand

cpp_ipc_remove <- function(id_sexp) {
  .Call(`_BiocParallel_cpp_ipc_remove`, id_sexp)
}

cpp_ipc_uuid <- function() {
  .Call(`_BiocParallel_cpp_ipc_uuid`)
}

cpp_ipc_locked <- function(id_sexp) {
  .Call(`_BiocParallel_cpp_ipc_locked`, id_sexp)
}

cpp_ipc_lock <- function(id_sexp) {
  .Call(`_BiocParallel_cpp_ipc_lock`, id_sexp)
}

cpp_ipc_try_lock <- function(id_sexp) {
  .Call(`_BiocParallel_cpp_ipc_try_lock`, id_sexp)
}

cpp_ipc_unlock <- function(id_sexp) {
  .Call(`_BiocParallel_cpp_ipc_unlock`, id_sexp)
}

cpp_ipc_value <- function(id_sexp) {
  .Call(`_BiocParallel_cpp_ipc_value`, id_sexp)
}

cpp_ipc_reset <- function(id_sexp, n) {
  .Call(`_BiocParallel_cpp_ipc_reset`, id_sexp, n)
}

cpp_ipc_yield <- function(id_sexp) {
  .Call(`_BiocParallel_cpp_ipc_yield`, id_sexp)
}
