\name{BiocParallel-deprecated}
\alias{BiocParallel-deprecated}
\title{Deprecated Functions in Package \sQuote{BiocParallel}}

\description{

  There are currently no deprecated functions in \sQuote{BiocParallel}.

}
